import { ApplicationRef, createComponent, inject, Injectable, inputBinding, } from "@angular/core";
import { jsPDF } from "jspdf";
import * as htmlToImage from "html-to-image";
import { Wait } from "@simplysm/sd-core-common";
import { SdBusyProvider } from "./sd-busy.provider";
import * as i0 from "@angular/core";
export class SdPrintProvider {
    #appRef = inject(ApplicationRef);
    #sdBusy = inject(SdBusyProvider);
    async printAsync(template, options) {
        await new Promise((resolve, reject) => {
            try {
                //-- busy
                this.#sdBusy.globalBusyCount.update(v => v + 1);
                //-- comp
                const compRef = createComponent(template.type, {
                    environmentInjector: this.#appRef.injector,
                    bindings: [
                        ...Object.keys(template.inputs).map((inputKey) => inputBinding(inputKey, () => template.inputs[inputKey])),
                    ],
                });
                const compEl = compRef.location.nativeElement;
                compEl.classList.add("_sd-print-template");
                //-- style
                const styleEl = document.createElement("style");
                styleEl.innerHTML = `
  @page {
      size: ${options?.size ?? "A4 auto"};
      margin: ${options?.margin ?? "0"};
  }
  body > ._sd-print-template { display: none; }
  @media print
  {
      html, body { -webkit-print-color-adjust: exact; background: white; }
      body > * { display: none !important; }
      body > ._sd-print-template { display: block !important; }
  }`;
                this.#appRef.attachView(compRef.hostView);
                //-- print
                document.body.appendChild(compEl);
                document.head.appendChild(styleEl);
                requestAnimationFrame(async () => {
                    try {
                        await Wait.until(() => compRef.instance.initialized());
                        await this.#waitForAllImagesLoadedAsync(compEl);
                        window.print();
                        resolve();
                    }
                    catch (err) {
                        reject(err);
                    }
                    finally {
                        try {
                            this.#sdBusy.globalBusyCount.update(v => v - 1);
                            styleEl.remove();
                            compRef.destroy();
                        }
                        catch { }
                    }
                });
            }
            catch (err) {
                reject(err);
            }
        });
    }
    async getPdfBufferAsync(template, options) {
        return await new Promise((resolve, reject) => {
            try {
                //-- busy
                this.#sdBusy.globalBusyCount.update(v => v + 1);
                //-- comp
                const compRef = createComponent(template.type, {
                    environmentInjector: this.#appRef.injector,
                    bindings: [
                        ...Object.keys(template.inputs).map((inputKey) => inputBinding(inputKey, () => template.inputs[inputKey])),
                    ],
                });
                const compEl = compRef.location.nativeElement;
                compEl.classList.add("_sd-print-template");
                //-- style
                const styleEl = document.createElement("style");
                styleEl.innerHTML = `html, body { overflow: hidden }`;
                this.#appRef.attachView(compRef.hostView);
                //-- print
                document.body.appendChild(compEl);
                document.head.appendChild(styleEl);
                requestAnimationFrame(async () => {
                    try {
                        await Wait.until(() => compRef.instance.initialized());
                        await this.#waitForAllImagesLoadedAsync(compEl);
                        const doc = new jsPDF(options?.orientation ?? "p", "pt", "a4");
                        doc.deletePage(1);
                        let els = compEl.findAll("sd-print-page");
                        els = els.length > 0 ? els : [compEl];
                        for (const el of els) {
                            const canvas = await htmlToImage.toCanvas(el, {
                                backgroundColor: "white",
                                pixelRatio: 4,
                            });
                            const orientation = el.getAttribute("sd-orientation");
                            doc.addPage("a4", orientation ?? "p").addImage({
                                imageData: canvas,
                                x: 0,
                                y: 0,
                                ...(orientation === "landscape"
                                    ? {
                                        height: 841.89,
                                        width: 595.28,
                                    }
                                    : {
                                        width: 595.28,
                                        height: 841.89,
                                    }),
                            });
                        }
                        const arrayBuffer = doc.output("arraybuffer");
                        resolve(Buffer.from(arrayBuffer));
                    }
                    catch (err) {
                        reject(err);
                    }
                    finally {
                        try {
                            this.#sdBusy.globalBusyCount.update(v => v - 1);
                            styleEl.remove();
                            compRef.destroy();
                        }
                        catch { }
                    }
                });
            }
            catch (err) {
                reject(err);
            }
        });
    }
    async #waitForAllImagesLoadedAsync(container) {
        const imgs = Array.from(container.querySelectorAll("img"));
        await Promise.all(imgs.map((img) => {
            return new Promise((resolve) => {
                if (img.complete && img.naturalWidth !== 0) {
                    resolve(); // 이미 로드된 이미지
                }
                else {
                    const onLoad = () => {
                        img.removeEventListener("load", onLoad);
                        img.removeEventListener("error", onLoad);
                        resolve();
                    };
                    img.addEventListener("load", onLoad);
                    img.addEventListener("error", onLoad); // 오류도 포함 (선택)
                }
            });
        }));
    }
    static { this.ɵfac = function SdPrintProvider_Factory(__ngFactoryType__) { return new (__ngFactoryType__ || SdPrintProvider)(); }; }
    static { this.ɵprov = /*@__PURE__*/ i0.ɵɵdefineInjectable({ token: SdPrintProvider, factory: SdPrintProvider.ɵfac, providedIn: "root" }); }
}
(() => { (typeof ngDevMode === "undefined" || ngDevMode) && i0.ɵsetClassMetadata(SdPrintProvider, [{
        type: Injectable,
        args: [{ providedIn: "root" }]
    }], null, null); })();
//# sourceMappingURL=sd-print.provider.js.map