import { ApplicationRef, createComponent, inject, Injectable, inputBinding, outputBinding, } from "@angular/core";
import { SdToastContainerControl } from "../controls/sd-toast-container.control";
import { SdSystemLogProvider } from "./sd-system-log.provider";
import { SdToastControl } from "../controls/sd-toast.control";
import { $signal } from "../utils/bindings/$signal";
import { $effect } from "../utils/bindings/$effect";
import * as i0 from "@angular/core";
export class SdToastProvider {
    constructor() {
        this.#appRef = inject(ApplicationRef);
        this.#systemLog = inject(SdSystemLogProvider);
        this.alertThemes = $signal([]);
        this.overlap = $signal(false);
    }
    #appRef;
    #systemLog;
    #containerRef;
    get containerRef() {
        if (this.#containerRef == null) {
            const compRef = createComponent(SdToastContainerControl, {
                environmentInjector: this.#appRef.injector,
                bindings: [inputBinding("overlap", this.overlap)],
            });
            document.body.appendChild(compRef.location.nativeElement);
            this.#appRef.attachView(compRef.hostView);
            this.#containerRef = compRef;
        }
        return this.#containerRef;
    }
    async try(fn, messageFn) {
        try {
            return await fn();
        }
        catch (err) {
            if (err instanceof Error) {
                if (messageFn) {
                    this.danger(messageFn(err));
                }
                else {
                    this.danger(err.message);
                }
                await this.#systemLog.writeAsync("error", err.stack);
                return undefined;
            }
            else {
                throw err;
            }
        }
    }
    notify(toast) {
        // container
        const containerEl = this.containerRef.location.nativeElement;
        // component
        const compRef = createComponent(toast.type, {
            environmentInjector: this.#appRef.injector,
            bindings: [
                ...Object.keys(toast.inputs).map((inputKey) => inputBinding(inputKey, () => toast.inputs[inputKey])),
                outputBinding("close", () => closeFn()),
            ],
        });
        // toast
        const bindings = { open: $signal(false) };
        const toastRef = createComponent(SdToastControl, {
            environmentInjector: this.#appRef.injector,
            projectableNodes: [[compRef.location.nativeElement]],
            bindings: [inputBinding("open", bindings.open)],
        });
        const toastEl = toastRef.location.nativeElement;
        containerEl.appendChild(toastEl);
        this.#appRef.attachView(toastRef.hostView);
        this.#appRef.attachView(compRef.hostView);
        bindings.open.set(true);
        window.setTimeout(() => {
            closeFn();
        }, 5000);
        function closeFn() {
            toastEl.addEventListener("transitionend", () => {
                compRef.destroy();
                toastRef.destroy();
            });
            bindings.open.set(false);
        }
        return bindings;
    }
    info(message, userProgress = false) {
        return this.#show("info", message, userProgress);
    }
    success(message, useProgress = false) {
        return this.#show("success", message, useProgress);
    }
    warning(message, useProgress = false) {
        return this.#show("warning", message, useProgress);
    }
    danger(message, useProgress = false) {
        return this.#show("danger", message, useProgress);
    }
    #show(theme, message, useProgress) {
        this.beforeShowFn?.(theme);
        if (this.alertThemes().includes(theme)) {
            alert(message);
            return undefined;
        }
        const bindings = {
            progress: $signal(0),
            message: $signal(message),
            open: $signal(false),
        };
        const toastRef = createComponent(SdToastControl, {
            environmentInjector: this.#appRef.injector,
            bindings: [
                inputBinding("useProgress", () => useProgress),
                inputBinding("progress", bindings.progress),
                inputBinding("message", bindings.message),
                inputBinding("theme", () => theme),
                inputBinding("open", bindings.open),
            ],
        });
        const toastEl = toastRef.location.nativeElement;
        const containerEl = this.containerRef.location.nativeElement;
        if (this.overlap()) {
            for (const child of Array.from(containerEl.children)) {
                containerEl.removeChild(child);
            }
        }
        containerEl.appendChild(toastEl);
        this.#appRef.attachView(toastRef.hostView);
        // repaint
        containerEl.repaint();
        bindings.open.set(true);
        if (useProgress) {
            $effect([bindings.progress], () => {
                if (bindings.progress() < 100)
                    return;
                closeAfterTime(1000);
            }, {
                injector: this.#appRef.injector,
            });
        }
        else {
            closeAfterTime(3000);
        }
        function closeAfterTime(ms) {
            window.setTimeout(() => {
                if (toastEl.matches(":hover")) {
                    closeAfterTime(1000);
                }
                else {
                    toastEl.addEventListener("transitionend", () => {
                        toastRef.destroy();
                    });
                    bindings.open.set(false);
                }
            }, ms);
        }
        return bindings;
    }
    static { this.ɵfac = function SdToastProvider_Factory(__ngFactoryType__) { return new (__ngFactoryType__ || SdToastProvider)(); }; }
    static { this.ɵprov = /*@__PURE__*/ i0.ɵɵdefineInjectable({ token: SdToastProvider, factory: SdToastProvider.ɵfac, providedIn: "root" }); }
}
(() => { (typeof ngDevMode === "undefined" || ngDevMode) && i0.ɵsetClassMetadata(SdToastProvider, [{
        type: Injectable,
        args: [{ providedIn: "root" }]
    }], null, null); })();
//# sourceMappingURL=sd-toast.provider.js.map